/*
								+---------------------------------+
								|                                 |
								|  ***   Tiny jpeg decoder   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Jpeg.h              |
								|                                 |
								+---------------------------------+
*/

#ifndef __BMP_INCLUDED
#define __BMP_INCLUDED

struct TBmp {
	char n_former_bpp; // former bpp, before conversion to RGBA8
	bool b_grayscale;
	bool b_alpha;
	int n_width, n_height;
	unsigned __int32 *p_buffer;
};

#endif //__BMP_INCLUDED

#ifndef __TINY_JPEG_INCLUDED
#define __TINY_JPEG_INCLUDED

/*
 *		=== config section ===
 */

//#define __VERBOSE
// if defined, print tables, header info, ... (to stdout)

//#define __DECODE_VERBOSE
// if defined, print decoding progress (i.e. huffman codes, DCT coeffs; to stdout)

#define __STRIP_ERROR_CHECKS
// if defined, all error checks aparts from alloc success tests are removed - saves a few code bytes
// (use with precaution - only if your set of images never changes, like in resource section)

/*
 *		=== ~config section ===
 */

class CJpegDecoder {
protected:
	static char p_zig_indices[/*64*/63]; // don't need the first one; actualy inverse zig sequence

	class CFileBuffer { // class for fast buffered file input (also assures byte-order reversion)
	protected:
		unsigned char m_p_buffer[16384];
		static const int m_n_buffer_size;
		int m_n_buffer_used;
		unsigned char *m_p_buffer_ptr;
		FILE *m_p_fr;

	public:
		CFileBuffer(FILE *p_fr);
		bool __fastcall Read_Byte(unsigned char &r_n_byte);
		bool __fastcall Read_Short(short &r_n_short);
		bool inline Read_Short(unsigned short &r_n_short) { return Read_Short((short&)r_n_short); }
		bool __fastcall Read_Marker(unsigned short &r_n_marker_code);
		bool __fastcall Skip_Bytes(int n_byte_num);
	};

	class CBitReader { // class for reading bit-by-bit, used in huffman decoder
	protected:
		unsigned char m_n_cur_byte; // current byte, it's shifted as we eat bits
		bool m_b_0xff_byte; // set to true if last byte read was 0xff
		unsigned char m_n_bit_num; // number of valid bits in byte
		bool m_b_error_flag;

	public:
		CBitReader();
		bool inline b_Error() { return m_b_error_flag; }
		unsigned char __fastcall n_GetBit(CFileBuffer *p_file_buffer);
		short __fastcall n_GetNumber(int n_bit_num, CFileBuffer *p_file_buffer);
		void Restart();
	};

	struct TBlockAction { // table for block actions
		unsigned short n_id;
		bool (*p_function)(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	};
	static const TBlockAction m_p_action_table[];

	struct TQuantTable { // quantization table
		bool b_16_bit; // otherwise 8-bit
		float p_value[64]; // hack! - should be unsigned short, but i pre-multiplied it and can use faster IDCT function
	} m_p_quant_table[4];
	// quantization tables (can be separate one for each component)

	struct TComponentInfo {
		unsigned char n_component_id; // 1 = Y, 2 = Cb, 3 = Cr (JFIF standard)
		unsigned char n_sampling_horiz;
		unsigned char n_sampling_vert; // component sampling (1, 2, or 4 usually)
		unsigned char n_scale_horiz;
		unsigned char n_scale_vert;
		unsigned char n_quant_table_index; // used quantization table
	};
	// outside TImageHeader because of scope issues (needed visibility for TScanHeader::TComponentInfo)

	struct TImageHeader {
		unsigned short n_frame_start_marker; // marker_SOF0_HuffBaseline_DCT supported for now only
		unsigned char n_sample_precision; // 8 (8 or 12 for extended baseline or progressive)
		short n_height;
		short n_width;
		int n_macro_block_num, n_macro_block_x_num; // number of macro-blocks
		unsigned char n_component_num; // should be 1-4 (1 = gray, 3 = YCbCr, 4 = CMYK)
		TComponentInfo p_component_info[4];
	} m_t_image_header;
	// image header

	struct TRestartInterval {
		bool b_enabled;
		unsigned short n_interval;
	} m_t_restart;
	// restart interval definition

	struct THuffmanTable {
		unsigned char n_code_num[16]; // number of codes for all possible 16 lengths
		unsigned char *p_code[16]; // 16 pointers into p_code_table where the actual data are
		unsigned short n_min_code_value[16]; // min and max code value; n+1 bit code must be >= n_min_code_value[n] && < n_max_code_value[n] (thus in case n_code_num[n] == 0 then also n_min_code_value[n] == n_max_code_value[n])
		//unsigned short n_max_code_value[16]; // don't store this, it's equal to n_min_code_value[n] + n_code_num[n]
		unsigned char p_code_table[256]; // table for all huffman codes
	} m_p_huffman_table[4][2]; // 0 = AC, 1 = DC
	// huffman tables (can be separate one for each component)

	struct TScanHeader {
		unsigned char n_component_num; // should match TImageHeader::n_component_num

		struct TComponentInfo2 {
			unsigned char n_component_id; // should match TImageHeader::TComponentInfo::n_component_id
			TComponentInfo *p_component; // points to image header component
			unsigned char n_DC_entropy_table;
			unsigned char n_AC_entropy_table; // table indices <0, 3>
			THuffmanTable *p_DC_table, *p_AC_table;
		} p_component_info[4];

		unsigned char n_spectrum_first; // index (in zig-zag order) of first coded coefficient
		unsigned char n_spectrum_last; // index (in zig-zag order) of last coded coefficient (should be set to 63 for sequential DCT)
		unsigned char n_sucessive_approximation; // two nibbles, point transform, should be zero (not used in marker_SOF0_HuffBaseline_DCT)
	};

	struct TImageData {
		bool b_read_image_header;
		unsigned char *p_framebuffer[4]; // frame-buffers, as big as image size / component sampling
		short n_framebuffer_width[4];
		short n_framebuffer_height[4];
		short n_dc_value[4]; // dc accumulators
	} m_t_image_data; // cleared always when starting decoding jpeg

public:
	TBmp *p_Decode_JPEG(FILE *p_fr);

protected:
	static bool Read_RestartInterval(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_SkipMarker(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_JFIF_Header(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_Unsupported(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_Skip(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_SkipShort(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_QuantTables(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_HuffmanTables(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_FrameHeader(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_StartScan(CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	bool Decode_MacroBlock(CBitReader &r_bit_reader, int n_macro_block_index,
		TScanHeader *p_scan_head, CFileBuffer *p_file_buffer);
};

#endif //__TINY_JPEG_INCLUDED
